@@ -882,7 +882,7 @@ int partition_copy_to_s3(ms3_st *s3_client, const char *aws_bucket,
   if ((error= s3_read_file_from_disk(filename, &alloc_block, &frm_length, 0)))
   {
     /*
-      In case of ADD PARTITION PARTITON the .frm file is already renamed.
+      In case of ADD PARTITION the .frm file is already renamed.
       Copy the renamed file if it exists.
     */
     fn_format(filename, path, "", ".frm", MY_REPLACE_EXT);
@@ -1198,7 +1198,7 @@ my_bool s3_rename_object(ms3_st *s3_client, const char *aws_bucket,
       if (!(errmsg= ms3_server_error(s3_client)))
         errmsg= ms3_error(error);
 
-      my_printf_error(EE_READ, "Got error from move_object(%s -> %s): %d %",
+      my_printf_error(EE_READ, "Got error from move_object(%s -> %s): %d %s",
                       error_flags,
                       from_name, to_name, error, errmsg);
     }
